require 'rulator/terms'

module Ruby
  # In the topmost Actions class event handlers should transform their input
  # array and pass it on. Example if on_if_mod which should pass on to its
  # mother on_if after having reformatted the results to fit its indata
  # expectations. This way users that inherit can limit the number of event
  # handlers they need to write if they really don't need to keep all cases
  # separated.

  class ParseActions
    def missing_method(id, *args)
      puts "Calling #{id.inspect} with args #{args.inspect}"
    end
  end

  # Default parse actions for Ruby grammar
  class AbstractSyntaxTreeBuilder < ParseActions
    # program : compstmt
    def on_compound_statement(results) 
      results[0]
    end

    # stmts : stmt
    def on_single_statement(results)   
      Body[results[0]] # A single statement is always wrapped up
    end

    # stmts : stmts terms stmt
    def on_another_statement(results)  
      results[0] = Body[] unless results[0]
      results[0].push(results[2])
      results[0]
    end

    # stmts : error stmt
    def on_statements_error(results)   
      puts "Warning! Parse error:\n #{results[0].inspect}"
      nil
    end

    # stmt : kALIAS fitem fitem
    #      | kALIAS tGVAR tGVAR
    #      | kALIAS tGVAR tSVAR
    #      | kALIAS tGVAR tBACK_REF
    def on_alias(results)
      Alias[results[1], results[2]]
    end

    # stmt : kUNDEF undef_list
    def on_undef(results)
      Undef[results[1]]
    end

    # stmt : stmt kIF_MOD expr
    def on_if_mod(results)
      If[results[2], results[0]]
    end

    # arg     : arg '?' arg ':' arg
    # if_tail : expr then compstmt
    def on_if(results)
      If[results[0], results[2], results[4]]
    end

    # stmt : stmt kUNLESS_MOD expr
    def on_unless_mod(results)
      Unless[results[2], results[0]]
    end

    # stmt : stmt kWHILE_MOD expr
    def on_while_mod(results)
      While[results[2], results[0]]
    end

    # stmt : stmt kUNTIL_MOD expr
    def on_until_mod(results)
      Until[results[2], results[0]]
    end

    # stmt : stmt kRESCUE_MOD stmt
    def on_rescue_mod(results)
      Rescue[results[0], results[2]]
    end

    # rescue : kRESCUE exc_list exc_var then compstmt rescue
    def on_rescue(results)
      res = Rescue[results[1], results[2], results[4]]
      # handle the chained ones
      res
    end

    
    # stmt : klBEGIN '{' compstmt '}'
    def on_begin_block(results)
      BeginBlock[results[2]]
    end

    # do_block : kDO_BLOCK opt_block_var compstmt kEND
    def on_block(results)
      Block[results[1], results[2]]
    end

    # stmt : klEND '{' compstmt '}'
    def on_end_block(results)
      EndBlock[results[2]]
    end

    # stmt : lhs '=' command_call
    #      | lhs '=' mrhs_basic
    # arg  : lhs '=' arg
    # stmt : mlhs '=' command_call
    #      | mlhs '=' mrhs
    def on_assignment(results)
      Assign[results[0], results[2]]
    end

    # expr : kRETURN call_args
    def on_return(results)
      Return[results[1]]
    end

    # expr : kBREAK call_args
    def on_break(results)
      Break[results[1]]
    end

    # primary : kRETRY
    def on_retry(results)
      Retry[]
    end

    # primary : kREDO
    def on_redo(results)
      Redo[]
    end

    # expr : kNEXT call_args
    def on_next(results)
      Next[results[1]]
    end

    # expr : expr kAND expr
    def on_and(results)
      And[results[0], results[2]]
    end

    # expr : expr kOR expr
    def on_or(results)
      Or[results[0], results[2]]
    end

    # expr : kNOT expr
    def on_not(results)
      Not[results[1]]
    end

    # expr : '!' command_call
    # arg  : '!'  arg
    def on_exclamation(results)
      Exclamation[results[1]]
    end

    # arg  : '~'  arg
    def on_invert_expr(results)
      Invert[results[1]]
    end

    # command : kSUPER command_args
    def on_super(results)
      Super[results[1]]
    end

    # command : kYIELD call_args
    def on_yield(results)
      Yield[results[1]]
    end

    def on_mlhs(results)
      Mlhs[results[0], results[1]]
    end

    # mlhs_node : primary '[' aref_args ']'
    # lhs       : primary '[' aref_args ']'
    def on_slice_var(results)
      SliceVar[results[0], results[2]]
    end

    # mlhs_node : primary '.'     tIDENTIFIER
    #           | primary tCOLON2 tIDENTIFIER
    #           | primary '.'     tCONSTANT
    # lhs       : primary '.'     tIDENTIFIER
    #           | primary tCOLON2 tIDENTIFIER
    #           | primary '.'     tCONSTANT
    def on_access(results)
      Access[results[0], results[2]]
    end

    # arg : variable tOP_ASGN arg
    def on_op_assign(results)
      # Not sure what to return here. Always OpAssign or maybe even more
      # fine details: PlusAssign etc? Need feedback?
      case results[1].info
      when :tOROP
	OrAssign[results[0], results[2]]
      when :tANDOP
	AndAssign[results[0], results[2]]
      else
	OpAssign[results[0], results[2], results[1].info]
      end
    end

    # arg : arg tDOT2 arg
    #     | arg tDOT3 arg
    def on_range(results)
      case results[1].token_type
      when :tDOT2
	ARange[results[0], results[2], true]
      when :tDOT3
	ARange[results[0], results[2], false]
      end
    end

    # arg : arg '+' arg
    def on_plus_expr(results)
      Plus[results[0], results[2]]
    end

    # arg : arg '-' arg
    def on_minus_expr(results)
      Minus[results[0], results[2]]
    end

    # arg : arg '*' arg
    def on_mult_expr(results)
      Mult[results[0], results[2]]
    end

    # arg : arg '/' arg
    def on_div_expr(results)
      Div[results[0], results[2]]
    end

    # arg : arg '%' arg
    def on_mod_expr(results)
      Mod[results[0], results[2]]
    end

    # arg : arg tPOW arg
    def on_pow_expr(results)
      Pow[results[0], results[2]]
    end

    # arg : tUPLUS arg
    def on_unaryplus_expr(results)
      UnaryPlus[results[1]]
    end

    # arg : tUMINUS arg
    def on_unaryminus_expr(results)
      UnaryMinus[results[1]]
    end

    # arg : arg '|' arg
    def on_or_expr(results)
      Or[results[0], results[2]]
    end

    # arg : arg '^' arg
    def on_exor_expr(results)
      ExOr[results[0], results[2]]
    end

    # arg : arg '&' arg
    def on_and_expr(results)
      And[results[0], results[2]]
    end

    # arg : arg tCMP arg
    def on_compare_expr(results)
      Compare[results[0], results[2]]
    end

    # mrhs       : arg
    # mrhs_basic : args  ','   tSTAR arg
    #            | {nil} {nil} tSTAR arg
    def on_mrhs(results)
      Mrhs[results[0], results[3]]
    end

    # primary : kBEGIN compstmt rescue opt_else ensure kEND
    def on_begin_end_block(results)
      BeginEndBlock[results[1], results[2], results[3], results[4]]
    end

    # arg     : kDEFINED opt_nl arg
    # primary : kDEFINED opt_nl '(' expr ')'    {but parens skipped}
    def on_defined(results)
      Defined[results[2]]
    end

    # primary : kUNLESS expr then compstmt opt_else kEND
    def on_unless(results)
      Unless[results[1], results[3], results[4]] 
    end

    # primary : kWHILE expr do compstmt kEND
    def on_while(results)
      While[results[1], results[3]]
    end

    # primary : kUNTIL expr do compstmt kEND
    def on_unless(results)
      Until[results[1], results[3]]
    end

    # primary : kCASE expr  opt_terms case_body kEND
    #         | kCASE {nil} opt_terms case_body kEND
    def on_case(results)
      Case[results[1], results[3]]
    end

    # case_body : kWHEN when_args then compstmt cases
    def on_when(results)
      When[results[1], results[3], results[4]]
    end

    # primary : kFOR block_var kIN expr do compstmt kEND
    def on_for(results)
      For[results[1], results[3], results[5]]
    end

    # primary : kCLASS cname superclass compstmt kEND
    def on_class(results)
      DefClass[results[1], results[2], results[3]]
    end

    # primary : kDEF singleton dot_or_colon fname f_arglist 
    #                               compstmt {rescue opt_else ensure} kEND
    def on_class_method(results)
      DefClassMethod[results[1], results[3], results[4], results[5],
	results[6]]
    end

    # primary : kCLASS tLSHFT expr term compstmt kEND
    def on_singleton_class(results)
      SingletonClass[results[2], results[4]]
    end

    # primary : kMODULE cname compstmt kEND
    def on_module(results)
      DefModule[results[1], results[2]]
    end

    # primary : kDEF fname f_arglist 
    #             {as one: compstmt rescue opt_else ensure} 
    #           kEND
    def on_method(results)
      DefMethod[results[1], results[2], results[3]]
    end

    # block_call    : block_call '.'     operation2 opt_paren_args
    #               | block_call tCOLON2 operation2 opt_paren_args
    # method_call   : {nil}      {nil}   operation  paren_args
    #               | primary    '.'     operation2 opt_paren_args
    #               | primary    tCOLON2 operation2 paren_args
    # command       : {nil}      {nil}   operation  command_args
    #               | primary    '.'     operation2 command_args
    #               | primary    tCOLON2 operation2 command_args
    # block_command : block_call '.'     operation2 command_args
    #               | block_call tCOLON2 operation2 command_args
    def on_call(results)
      Call[results[0], results[2], results[3]]
    end

    # primary : method_call brace_block
    def on_call_with_block(results)
      unless results[0].arguments
	results[0].arguments = Args[nil, nil, nil, results[1]]
      else
	results[0].arguments.block = results[1]
      end
      results[0]
    end

    # method_call : primary tCOLON2 operation3
    def on_constant(results)
      # Should this really be its own term?
      ConstantAccess[results[0], results[2]] 
    end

    # method_call : kSUPER paren_args
    def on_super_call(results)
      Super[results[1]]
    end

    # string : string tSTRING
    #        | string tDSTRING
    def on_strings(results)
      # Concatenate them!
    end

    # f_args : f_arg ','   f_optarg ','   f_rest_arg opt_f_block_arg
    #        | f_arg ','   f_optarg {nil} {nil}      opt_f_block_arg
    #        | f_arg ','   {nil}    {nil} f_rest_arg opt_f_block_arg
    #        | f_arg {nil} {nil}    {nil} {nil}      opt_f_block_arg
    #        | {nil} {nil} f_optarg ','   f_rest_arg opt_f_block_arg
    #        | {nil} {nil} f_optarg {nil} {nil}      opt_f_block_arg
    #        | {nil} {nil} {nil}    {nil} f_rest_arg opt_f_block_arg
    #        | {nil} {nil} {nil}    {nil} {nil}      f_block_arg
    def on_formal_args(results)
      Args[results[0], results[2], results[4], results[5]]
    end

    # call_args  : command {nil} {nil}  {nil} {nil} {nil} {nil}
    #            | args    ','   assocs ','   tSTAR arg   opt_block_arg
    #            | args    ','   assocs {nil} {nil} {nil} opt_block_arg
    #            | args    ','   {nil}  {nil} tSTAR arg   opt_block_arg
    #            | args    {nil} {nil}  {nil} {nil} {nil} opt_block_arg
    #            | {nil}   {nil} assocs ','   tSTAR arg   opt_block_arg
    #            | {nil}   {nil} assocs {nil} {nil} {nil} opt_block_arg
    #            | {nil}   {nil} {nil}  {nil} tSTAR arg   opt_block_arg
    #            | {nil}   {nil} {nil}  {nil} {nil} {nil} block_arg
    def on_args(results)
      Args[results[0], results[2], results[5], results[6]]
    end

    # symbol : tSYMBEG sym
    def on_symbol(results)
      SymbolLit[results[1].lexeme]
    end

    def keyword_to_token(keyword, token)
      t = token[]
      t.position = keyword.position
      t
    end

    def on_true(results)
      keyword_to_token(results[0], True)
    end

    def on_false(results)
      keyword_to_token(results[0], False)
    end

    def on_self(results)
      keyword_to_token(results[0], Self)
    end

    def on_nil(results)
      keyword_to_token(results[0], Nil)
    end

    # primary : tLBRACK aref_args ']'
    def on_array(results)
      ArrayLit[*results[1]]
    end

    # primary : tLBRACE assoc_list '}'
    def on_hash(results)
      # Should we use a hash? Probably not since we'd lose the ordering...
      HashLit[*results[1]] 
    end

    # assoc : arg tASSOC arg
    def on_assoc(results)
      Assoc[results[0], results[2]]
    end

    # arg     : {nil}  arg '?'   arg      ':'   arg
    # if_tail : kELSIF expr then compstmt {nil} if_tail
    # primary : kIF    expr then compstmt {nil} if_tail kEND
    def on_if(results)
      If[results[1], results[3], results[5]]
    end

    # primary : kUNLESS expr then compstmt {nil} if_tail kEND
    def on_unless(results)
      Unless[results[1], results[3], results[5]]
    end

    # arg : arg '<' arg
    def on_lessthan_expr(results)
      LessThan[results[0], results[2]]
    end

    # arg : arg '>' arg
    def on_greaterthan_expr(results)
      GreaterThan[results[0], results[2]]
    end

    # arg : arg tLEQ arg
    def on_leq_expr(results)
      LessThanEq[results[0], results[2]]
    end

    # arg : arg tGEQ arg
    def on_geq_expr(results)
      GreaterThanEq[results[0], results[2]]
    end

    # arg : arg tEQ arg
    def on_eq_expr(results)
      Eq[results[0], results[2]]
    end

    # arg : arg tNEQ arg
    def on_neq_expr(results)
      NotEq[results[0], results[2]]
    end

    # arg : arg tMATCH arg
    def on_match_expr(results)
      Match[results[0], results[2]]
    end

    # arg : arg tNMATCH arg
    def on_notmatch_expr(results)
      NotMatch[results[0], results[2]]
    end

    # arg : arg tEQQ arg
    def on_eqq_expr(results)
      Eqq[results[0], results[2]]
    end

    # arg : arg tLSHFT arg
    def on_lshift_expr(results)
      LeftShift[results[0], results[2]]
    end

    # arg : arg tRSHFT arg
    def on_rshift_expr(results)
      RightShift[results[0], results[2]]
    end

    # arg : arg tANDOP arg
    def on_cond_and_expr(results)
      CondAnd[results[0], results[2]]
    end

    # arg : arg tOROP arg
    def on_cond_or_expr(results)
      CondOr[results[0], results[2]]
    end

    def on_dstring(results)
      t = DStringLit[results[0].lexeme]
      t.position = results[0].position
      t
    end

    def on_string(results)
      t = StringLit[results[0].lexeme]
      t.position = results[0].position
      t
    end
  end # class ParseActions
end # module Ruby
