require 'rockit/term/term'

module Ruby
  # A Ruby term is a term with additional attributes:
  #    position      - position of first token that gave rise to term
  #    pre_comments  - array of comments (String) prior to the first token
  #    post_commetns - array of comments (String) after the last token
  #
  # Note that when we refer to the first and last tokens we mean the first
  # and last tokens in the production that gives rise to the term even though
  # all of them might not be part of the term.
  #
  # Example:
  #          # My class
  #          class T; def m; 1; end; end
  #
  #     The first token of the Class term is "class" even though the first
  #     childrne of the term is Identifier["T"]. In this example the 
  #     pre_comments of the Class term would be ["# My class"] while the 
  #     post_comments would be nil.
  #
  class Term < Rockit::Term
    extend Rockit::PartialEvalSubClassFactory
    attr_accessor :pre_comments, :post_comments, :position
  end

  # This is not nice but I don't see better approach yet.
  class ArrayTerm < Term
    extend Rockit::PartialEvalSubClassFactory
    def push(child)
      @children.push child
    end
  end

  # Tokens produced in the lexer
  Token       = Term.new_sub_class(:Token, :lexeme, :info)
  
  def self.token(name, tokenType)
    t = Token.new_sub_class(name)
    # FIXME: Do this with class instance vars instead!
    t.class_eval <<-EOS
      def token_type
	#{tokenType.inspect}
      end
    EOS
    t
  end

  # Since the parser require a symbol we use the default Ruby symbols
  # to set the token_type.
  GlobalVar   = token(:GlobalVar,   :tGVAR)
  InstanceVar = token(:InstanceVar, :tIVAR)
  ClassVar    = token(:ClassVar,    :tCVAR)
  BackRef     = token(:BackRef,     :tBACK_REF)
  NthRef      = token(:NthRef,      :tNTH_REF)
  Identifier  = token(:Identifier,  :tIDENTIFIER)
  Constant    = token(:Constant,    :tCONSTANT)
  IntegerLit  = token(:IntegerLit,  :tINTEGER)
  FloatLit    = token(:FloatLit,    :tFLOAT)

  # Terms produced in the parser.
  Body        = ArrayTerm.new_sub_class(:Body)
  Alias       = Term.new_sub_class(:Alias,       :old, :new)
  If          = Term.new_sub_class(:If,          :condition, :then, :else)
  Unless      = Term.new_sub_class(:Unless,      :condition, :then, :else)
  Return      = Term.new_sub_class(:Return,      :arguments)
  Break       = Term.new_sub_class(:Break,       :arguments)
  Next        = Term.new_sub_class(:Next,        :arguments)

  Expr        = Term.new_sub_class(:Expr)
  And         = Expr.new_sub_class(:And,         :left, :right)
  Or          = Expr.new_sub_class(:Or,          :left, :right)
  Not         = Expr.new_sub_class(:Not,         :expr)
  Invert      = Expr.new_sub_class(:Invert,   :expr)
  Exclamation = Expr.new_sub_class(:Exclamation, :command)

  # Comparators
  Comparator  = Term.new_sub_class(:Comparator)
  Eq          = Comparator.new_sub_class(:Eq,         :left, :right)
  Eqq         = Comparator.new_sub_class(:Eqq,        :left, :right)
  Compare     = Comparator.new_sub_class(:Compare,        :left, :right)
  NotEq       = Comparator.new_sub_class(:NotEq,      :left, :right)
  Match       = Comparator.new_sub_class(:Match,      :left, :right)
  NotMatch    = Comparator.new_sub_class(:NotMatch,   :left, :right)
  LessThan    = Comparator.new_sub_class(:LessThan,   :left, :right)
  LessThanEq  = Comparator.new_sub_class(:LessThanEq, :left, :right)
  GreaterThan = Comparator.new_sub_class(:GreaterThan,   :left, :right)
  GreaterThanEq = Comparator.new_sub_class(:GreaterThanEq, :left, :right)

  CondAnd     = Term.new_sub_class(:CondAnd, :left, :right)
  CondOr      = Term.new_sub_class(:CondOr, :left, :right)

  DStringLit  = Term.new_sub_class(:DStringLit, :value)
  StringLit   = Term.new_sub_class(:StringLit, :value)

  Super       = Term.new_sub_class(:Super,       :arguments)
  Yield       = Term.new_sub_class(:Yield,       :arguments)

  True        = token(:True,  :kTRUE)
  False       = token(:False, :kFALSE)
  Nil         = token(:Nil,   :kNIL)
  Self        = token(:Self,  :kSELF)

  ArrayLit    = Term.new_sub_class(:ArrayLit, :elements, :rest)
  HashLit     = ArrayTerm.new_sub_class(:HashLit)
  Assoc       = Term.new_sub_class(:Assoc,  :from, :to)

  SliceVar    = Term.new_sub_class(:SliceVar,    :primary, :args)
  Access      = Term.new_sub_class(:Access,      :primary, :attribute)

  Args        = Term.new_sub_class(:Args,        :args, :opt, :rest, :block)
  DefMethod   = Term.new_sub_class(:DefMethod,   :name, :args, :body)
  DefClassMethod = Term.new_sub_class(:DefClassMethod, :klass, :name, :args, 
				      :body, :rescue)
  DefClass    = Term.new_sub_class(:DefClass,    :name, :super, :body)
  DefModule   = Term.new_sub_class(:DefModule,   :name, :body)
  Call        = Term.new_sub_class(:Call,        :receiver, :message, 
				                            :arguments)
  Block       = Term.new_sub_class(:Block,       :args, :body)

  CondAssign  = Term.new_sub_class(:CondAssign,  :variable, :argument)
  OrAssign    = CondAssign.new_sub_class(:OrAssign)
  AndAssign   = CondAssign.new_sub_class(:AndAssign)
  OpAssign    = Term.new_sub_class(:OpAssign,  
				   :variable, :argument, :operator)

  Plus        = Term.new_sub_class(:Plus,        :left, :right)
  Minus       = Term.new_sub_class(:Minus,       :left, :right)
  Mult        = Term.new_sub_class(:Mult,        :left, :right)
  Div         = Term.new_sub_class(:Div,         :left, :right)
  Mod         = Term.new_sub_class(:Mod,         :left, :right)
  Pow         = Term.new_sub_class(:Pow,         :base, :exponent)
  UnaryPlus   = Term.new_sub_class(:UnaryPlus,   :val)
  UnaryMinus  = Term.new_sub_class(:UnaryMinus,  :val)
  ExOr        = Term.new_sub_class(:ExOr,        :left, :right)
  LeftShift   = Term.new_sub_class(:LeftShift,   :left, :right)
  RightShift  = Term.new_sub_class(:RightShift,   :left, :right)

  ARange       = Term.new_sub_class(:ARange, :begin, :end, :inclusive)

  SymbolLit   = Term.new_sub_class(:SymbolLit,   :lexeme)

  Loop        = Term.new_sub_class(:Loop)
  While       = Loop.new_sub_class(:While, :condition, :body)
  Until       = Loop.new_sub_class(:Until, :condition, :body)
  For         = Loop.new_sub_class(:For, :variable, :expr, :body)
  
  When        = Term.new_sub_class(:When, :args, :body, :next)
  Case        = Term.new_sub_class(:Case, :expr, :clauses)

  BeginBlock  = Term.new_sub_class(:BeginBlock,  :body)
  EndBlock    = Term.new_sub_class(:EndBlock,    :body)

  Assign      = Term.new_sub_class(:Assign,      :lhs,  :rhs)
  Mrhs        = Term.new_sub_class(:Mrhs,        :args, :rest)
  Mlhs        = Term.new_sub_class(:Mrhs,        :args, :rest)
end
