# Unit tests for Ruby lexer in rulator/lexer
# Written by Robert Feldt and Rich Kilmer
#
# Copyright (c) 2001, Robert Feldt, feldt@ce.chalmers.se
#
# $Id: utest_lexer.rb,v 1.19 2001/12/12 19:18:38 feldt Exp $
#
require 'rubyunit'
require 'rulator/lexer'
include Ruby

class RubyLexerTest < TestCase
	
  #---------------------- Individual Token Tests ---------------------------

  def test_01_initialize
    t = Lexer.new ""
    assert_kind_of(Lexer, t)
  end

  def test_02_end_of_stream
    assert_tok("", false, false)
  end
	
  def test_03_whitespace
    assert_tok(" 1",         :tINTEGER, 1)
    assert_tok("\t 1",       :tINTEGER, 1)
    assert_tok("\f \t1",     :tINTEGER, 1)
    assert_tok("\r \t\f1",   :tINTEGER, 1)
    assert_tok("\v \t\f\r1", :tINTEGER, 1)
    assert_tok(" \13a",      :tIDENTIFIER, "a") # RACK
  end

  def test_04_line_comment 
    # This is a deviation from Matz yylex which simply skips comments.
    # We can't do that since people will want to be notified of comments
    # to extract info from them, colour them etc.
    assert_tok("# A comment\n", :tComment, "# A comment", :postst => EXPR_BEG)

    assert_tok("# A comment", :tComment, "# A comment", :postst => EXPR_BEG)
  end

  def test_05_block_comment # =begin...=end
    rd_doc = <<-EOS
=begin
     Some crazy stuf...!
 test
=end
    EOS
    assert_tok(rd_doc, :tBlockComment, "     Some crazy stuf...!\n test\n")
  end

  def test_06_newline
    assert_tok("\n1", :tINTEGER, 1, 
	       :prest => [EXPR_BEG, EXPR_FNAME, EXPR_DOT]) 

    assert_tok("\na", "\n", "\n", 
	       :prest => [EXPR_END, EXPR_ARG, EXPR_CLASS, EXPR_MID],
	       :postst => EXPR_BEG) 
  end

  def test_07_star # *
    assert_tok("**=", :tOP_ASGN, "**=", :postst => EXPR_BEG, :info => :tPOW)
    assert_tok("**", :tPOW, "**", :postst => EXPR_BEG)
    assert_tok("*=", :tOP_ASGN, "*=", :postst => EXPR_BEG, :info => "*")
    assert_tok("*", :tSTAR, "*", :prest => EXPR_BEG, :postst => EXPR_BEG)
    assert_tok("*", :tSTAR, "*", :prest => EXPR_MID, :postst => EXPR_BEG)
    assert_tok("*", "*", "*", :prest => (EXPR_STATES - [EXPR_BEG, EXPR_MID]), 
	       :postst => EXPR_BEG)
  end

  def test_08_exclamation # !
    assert_tok("!=", :tNEQ, "!=", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("!~", :tNMATCH, "!~", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("!", "!", "!", :prest => EXPR_STATES, :postst => EXPR_BEG)
  end

  def test_09_equal_sign # =
    assert_tok("===", :tEQQ, "===", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("==",  :tEQ,  "==", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("=~",  :tMATCH,  "=~", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("=>",  :tASSOC,  "=>", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("=",  "=",  "=", :prest => EXPR_STATES, :postst => EXPR_BEG)    
  end

  def test_10_less_than
    assert_tok("<=>", :tCMP, "<=>", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("<=", :tLEQ, "<=", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("<<=", :tOP_ASGN, "<<=", :info => :tLSHFT,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("<<", :tLSHFT, "<<", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    assert_tok("<", "<", "<", :prest => EXPR_STATES, :postst => EXPR_BEG)
  end

  def test_11_greater_than
    assert_tok(">=", :tGEQ, ">=", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(">>=", :tOP_ASGN, ">>=", :info => :tRSHFT,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(">>", :tRSHFT, ">>", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(">", ">", ">", :prest => EXPR_STATES, :postst => EXPR_BEG)
  end

  def test_12_question_mark
    assert_tok("? ", "?", "?", :prest => END_AND_ENDARG, :postst => EXPR_BEG)

    assert_tok("?a", :tINTEGER, ?a, :prest => EXPR_STATES - END_AND_ENDARG,
	       :postst => EXPR_END)
    assert_tok("??", :tINTEGER, ??, :prest => EXPR_STATES - END_AND_ENDARG,
	       :postst => EXPR_END)

    assert_tok("?\\xa", :tINTEGER, ?\xa, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    assert_tok("?\\x01", :tINTEGER, ?\x01, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)

    assert_tok("?\\1", :tINTEGER, ?\1, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    assert_tok("?\\72", :tINTEGER, ?\72, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    assert_tok("?\\345", :tINTEGER, ?\345, 
	       :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)

    "ntrfvaebs".each_byte do |b|
      s = "?\\X"; s[2] = b      
      assert_tok(s, :tINTEGER, eval(s),
		 :prest => EXPR_STATES - END_AND_ENDARG, :postst => EXPR_END)
    end
  end

  def test_13_ampersand
    assert_tok("&&=", :tOP_ASGN, "&&=", :info => :tANDOP,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("&&", :tANDOP, "&&", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("&=", :tOP_ASGN, "&=", :info => "&",
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok(" &a", :tAMPER, "&", :prest => EXPR_ARG, :postst => EXPR_BEG) 
    assert_tok("&a", "&", "&", :prest => EXPR_ARG, :postst => EXPR_BEG) 
    assert_tok("&", :tAMPER, "&", :prest => [EXPR_BEG, EXPR_MID],
	       :postst => EXPR_BEG) 
    assert_tok("&", "&", "&", 
	       :prest => EXPR_STATES - [EXPR_BEG, EXPR_MID, EXPR_ARG],
	       :postst => EXPR_BEG) 
  end

  def test_14_bar
    assert_tok("||=", :tOP_ASGN, "||=", :info => :tOROP,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("||", :tOROP, "||", :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("|=", :tOP_ASGN, "|=", :info => "|",
	       :prest => EXPR_STATES, :postst => EXPR_BEG)
    assert_tok("|", "|", "|", :prest => EXPR_STATES, :postst => EXPR_BEG) 
  end

  def test_15_plus
    assert_tok("+@", :tUPLUS, "+@", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("+ ", "+", "+", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("+=", :tOP_ASGN, "+=", :info => "+",
	       :prest => EXPR_STATES - [EXPR_FNAME, EXPR_DOT], 
	       :postst => EXPR_BEG)
    assert_tok("+12345", :tINTEGER, +12345, :prest => BEG_AND_MID,
	       :postst => EXPR_END)
    assert_tok(" +12345", :tINTEGER, +12345, :prest => EXPR_ARG)
    assert_tok("+ ", :tUPLUS, "+", :prest => BEG_AND_MID, :postst => EXPR_BEG)
    #assert_tok(" + ", :tUPLUS, "+", :prest => EXPR_ARG, :postst => EXPR_BEG)
    assert_tok("+", "+", "+", 
	       :prest => EXPR_STATES - BEG_AND_MID - FNAME_AND_DOT -
	                 [EXPR_ARG],
	       :postst => EXPR_BEG)
  end

  def test_16_minus
    assert_tok("-@", :tUMINUS, "-@", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("- ", "-", "-", :prest => [EXPR_FNAME, EXPR_DOT])
    assert_tok("-=", :tOP_ASGN, "-=", :info => "-",
	       :prest => EXPR_STATES - [EXPR_FNAME, EXPR_DOT], 
	       :postst => EXPR_BEG)
    assert_tok("-12345", :tINTEGER, -12345, :prest => BEG_AND_MID,
	       :postst => EXPR_END)
    assert_tok(" -12345", :tINTEGER, -12345, :prest => EXPR_ARG, 
	       :postst => EXPR_END)
    assert_tok("- ", :tUMINUS, "-", :prest => BEG_AND_MID, :postst => EXPR_BEG)
    #assert_tok(" - ", :tUMINUS, "-", :prest => EXPR_ARG, :postst => EXPR_BEG)
    assert_tok("-", "-", "-", 
	       :prest => EXPR_STATES - BEG_AND_MID - FNAME_AND_DOT -
	                 [EXPR_ARG],
	       :postst => EXPR_BEG)
  end

  def test_17_dot
    assert_tok("...", :tDOT3, "...", :prest => EXPR_STATES, :postst => EXPR_BEG) 
    assert_tok("..", :tDOT2, "..", :prest => EXPR_STATES, :postst => EXPR_BEG) 
    assert_tok(".a", ".", ".", :prest => EXPR_STATES, :postst => EXPR_DOT)
    assert_tok(".5", :tFLOAT, 0.5)
  end

  def test_18_number_integer 
    assert_tok("2", :tINTEGER, 2)
    assert_tok("12", :tINTEGER, 12)
    assert_tok("-15", :tINTEGER, -15)
    assert_tok("12_000", :tINTEGER, 12000)
    assert_tok("987_6_5_432_10", :tINTEGER, 987_6_5_432_10)
  end
	
  def test_19_number_hex 
    assert_tok("0xf", :tINTEGER, 15)
    assert_tok("0xab", :tINTEGER, 10*16+11)
    assert_tok("0x__c_d_e", :tINTEGER, 12*16*16+13*16+14)
    assert_tok("0xf", :tINTEGER, 0xf)
    assert_tok("0xab", :tINTEGER, 0xa_b)
    assert_tok("0x__c_d_e", :tINTEGER, 0x__c_d_e)
  end
	
  def test_20_number_binary 
    assert_tok("0b01", :tINTEGER, 0b01)
    assert_tok("0b_11_0", :tINTEGER, 0b_11_0)
  end
	
  def test_21_number_octal 
    assert_tok("0123", :tINTEGER, 0123)
    assert_tok("0_7_6_5", :tINTEGER, 0_7_6_5)
  end

  def test_22_number_float 
    assert_tok("1.2345", :tFLOAT, 1.2345)
    assert_tok("-1e10", :tFLOAT, -1e10)
    assert_tok("-9.87E-1", :tFLOAT, -9.87E-1)
    # This one is not correct since it should be UMINUS followed by the float.
    # assert_tok("-.2345", :tFLOAT, -.2345)
  end

  def test_23_colon
    assert_tok(":test", :tSYMBEG, ":")
    assert_tok(": ", ":", ":", :postst => EXPR_BEG)
    assert_tok(":",  ":", ":", :prest => EXPR_END, :postst => EXPR_BEG)
    assert_tok("::b", :tCOLON2, "::", :prest => NOT_BEG_MID_OR_ARG, 
	       :postst => EXPR_DOT)
    assert_tok("::", :tCOLON3, "::", :prest => [EXPR_BEG, EXPR_MID], 
	       :postst => EXPR_BEG)
    assert_tok(" ::", :tCOLON3, "::", 
	       :prest => EXPR_ARG, :postst => EXPR_BEG)
  end

  def test_24_identifier
    assert_tok("my_ident", :tIDENTIFIER, "my_ident")
  end

  def test_25_constant
    assert_tok("My_ident_", :tCONSTANT, "My_ident_")    
  end

  def test_26_fid
    assert_tok("a?",  :tFID, "a?")
    assert_tok("z!",  :tFID, "z!")
  end

  def test_27_instance_variable
    assert_tok("@a",  :tIVAR, "a")
    assert_tok("@z_", :tIVAR, "z_")
  end

  def test_28_class_variable
    assert_tok("@@a",  :tCVAR, "a")
    assert_tok("@@z_", :tCVAR, "z_")
  end

  def test_29_global_variable
    assert_tok("$a",   :tGVAR, "a", :postst => EXPR_END)
    assert_tok("$z_",  :tGVAR, "z_", :postst => EXPR_END)
    assert_tok("$_za", :tGVAR, "_za", :postst => EXPR_END)

    assert_tok("$*",  :tGVAR, "*", :postst => EXPR_END)
    assert_tok("$$",  :tGVAR, "$", :postst => EXPR_END)
    assert_tok("$?",  :tGVAR, "?", :postst => EXPR_END)
    assert_tok("$!",  :tGVAR, "!", :postst => EXPR_END)
    assert_tok("$@",  :tGVAR, "@", :postst => EXPR_END)
    assert_tok("$/",  :tGVAR, "/", :postst => EXPR_END)
    assert_tok("$\\", :tGVAR, "\\", :postst => EXPR_END)
    assert_tok("$;",  :tGVAR, ";", :postst => EXPR_END)
    assert_tok("$,",  :tGVAR, ",", :postst => EXPR_END)
    assert_tok("$.",  :tGVAR, ".", :postst => EXPR_END)
    assert_tok("$=",  :tGVAR, "=", :postst => EXPR_END)
    assert_tok("$:",  :tGVAR, ":", :postst => EXPR_END)
    assert_tok("$<",  :tGVAR, "<", :postst => EXPR_END)
    assert_tok("$>",  :tGVAR, ">", :postst => EXPR_END)
    assert_tok("$\"", :tGVAR, "\"", :postst => EXPR_END)
  end

  def test_30_back_ref
    assert_tok("$&",   :tBACK_REF, "&", :postst => EXPR_END)
    assert_tok("$`",   :tBACK_REF, "`", :postst => EXPR_END)
    assert_tok("$'",   :tBACK_REF, "'", :postst => EXPR_END)
    assert_tok("$+",   :tBACK_REF, "+", :postst => EXPR_END)
  end

  def test_31_nth_ref
    assert_tok("$1",    :tNTH_REF, 1, :postst => EXPR_END)  
    assert_tok("$9999", :tNTH_REF, 9999, :postst => EXPR_END) 
  end

  def test_32_reserved_word
    # Positive tests
    assert_tok("end",      :kEND,        "end",      :postst => EXPR_END)    
    assert_tok("else",     :kELSE,       "else",     :postst => EXPR_BEG)    
    assert_tok("case",     :kCASE,       "case",     :postst => EXPR_BEG)    
    assert_tok("ensure",   :kENSURE,     "ensure",   :postst => EXPR_BEG)    
    assert_tok("module",   :kMODULE,     "module",   :postst => EXPR_BEG)    
    assert_tok("elsif",    :kELSIF,      "elsif",    :postst => EXPR_BEG)    
    assert_tok("def",      :kDEF,        "def",      :postst => EXPR_FNAME)    
    assert_tok("rescue",   :kRESCUE,     "rescue",   
	       :prest => EXPR_BEG, :postst => EXPR_MID)    
    assert_tok("rescue",   :kRESCUE_MOD, "rescue",   
	       :prest => EXPR_STATES - [EXPR_BEG], :postst => EXPR_BEG)    
    assert_tok("not",      :kNOT,        "not",      :postst => EXPR_BEG)    
    assert_tok("then",     :kTHEN,       "then",     :postst => EXPR_BEG)    
    assert_tok("yield",    :kYIELD,      "yield",    :postst => EXPR_ARG)    
    assert_tok("for",      :kFOR,        "for",      :postst => EXPR_BEG)    
    assert_tok("self",     :kSELF,       "self",     :postst => EXPR_END)    
    assert_tok("false",    :kFALSE,      "false",    :postst => EXPR_END)    
    assert_tok("retry",    :kRETRY,      "retry",    :postst => EXPR_END)    
    assert_tok("return",   :kRETURN,     "return",   :postst => EXPR_MID)    
    assert_tok("true",     :kTRUE,       "true",     :postst => EXPR_END)    
    assert_tok("if",       :kIF,         "if",       :postst => EXPR_BEG)    
    assert_tok("if",       :kIF_MOD,     "if",   :prest => EXPR_END, :postst => EXPR_BEG)    
    #assert_tok("defined?", :kDEFINED,    "defined?", :postst => EXPR_ARG)    
    assert_tok("super",    :kSUPER,      "super",    :postst => EXPR_ARG)    
    assert_tok("undef",    :kUNDEF,      "undef",    :postst => EXPR_FNAME)    
    assert_tok("break",    :kBREAK,      "break",    :postst => EXPR_END)    
    assert_tok("in",       :kIN,         "in",       :postst => EXPR_BEG)    
    assert_tok("do",       :kDO,         "do",       :postst => EXPR_BEG)    
    #assert_tok("do",       :kDO_COND,    "do",  :postst => EXPR_BEG)  {|t| t.cond_push}
    #assert_tok("do",       :kDO_BLOCK,   "do",  :postst => EXPR_BEG)  {|t| t.cmdarg_push}
    assert_tok("nil",      :kNIL,        "nil",      :postst => EXPR_END)    
    assert_tok("until",    :kUNTIL,      "until",    :postst => EXPR_BEG)
    assert_tok("until",    :kUNTIL_MOD,  "until",    :prest => EXPR_END, :postst => EXPR_BEG)    
    assert_tok("unless",   :kUNLESS,     "unless",   :postst => EXPR_BEG)    
    assert_tok("unless",   :kUNLESS_MOD, "unless",  :prest => EXPR_END,   :postst => EXPR_BEG)    
    assert_tok("or",       :kOR,         "or",       :postst => EXPR_BEG)    
    assert_tok("next",     :kNEXT,       "next",     :postst => EXPR_END)    
    assert_tok("when",     :kWHEN,       "when",     :postst => EXPR_BEG)    
    assert_tok("redo",     :kREDO,       "redo",     :postst => EXPR_END)    
    assert_tok("and",      :kAND,        "and",      :postst => EXPR_BEG)    
    assert_tok("begin",    :kBEGIN,      "begin",    :postst => EXPR_BEG)    
    assert_tok("__LINE__", :k__LINE__,   "__LINE__", :postst => EXPR_END)    
    assert_tok("__FILE__", :k__FILE__,   "__FILE__", :postst => EXPR_END)    
    assert_tok("class",    :kCLASS,      "class",    :postst => EXPR_CLASS)    
    assert_tok("END",      :klEND,       "END",      :postst => EXPR_END)    
    assert_tok("BEGIN",    :klBEGIN,     "BEGIN",    :postst => EXPR_END)    
    assert_tok("while",    :kWHILE,      "while",    :postst => EXPR_BEG)    
    assert_tok("alias",    :kALIAS,      "alias",    :postst => EXPR_FNAME)    

    # Negative tests
    assert_tok("enda",     :tIDENTIFIER,  "enda")
    assert_tok("alias1",   :tIDENTIFIER,  "alias1")
    assert_tok("and?",     :tFID,         "and?")
    assert_tok("BEGINA",   :tCONSTANT,    "BEGINA")
    assert_tok("__self__", :tIDENTIFIER,  "__self__")
  end
  
  def test_33_comma
    assert_tok(",", ",", ",", :prest => EXPR_STATES,
	       :postst => EXPR_BEG)
  end

  def test_34_semicolon
    assert_tok(";", ";", ";", :prest => EXPR_STATES,
	       :postst => EXPR_BEG)
  end

  def test_35_caret
    assert_tok("^=", :tOP_ASGN, "^=", :info => "^",
	       :prest => EXPR_STATES, :postst => EXPR_BEG)    
    assert_tok("^", "^", "^", :prest => EXPR_STATES,
	       :postst => EXPR_BEG)    
  end

  def test_36_tilde
    assert_tok("~@", "~", "~", #:postpos => 2, 
	       :prest => [EXPR_FNAME, EXPR_DOT], :postst => EXPR_BEG)    
    assert_tok("~", "~", "~", :postpos => 1,
	       :prest => EXPR_STATES, :postst => EXPR_BEG)    
  end

  def test_37_slash
    assert_tok("/=", :tOP_ASGN, "/",
	       :prest => EXPR_STATES - [EXPR_BEG, EXPR_MID],
	       :postst => EXPR_BEG)
    assert_tok("/", "/", "/", :prest => EXPR_STATES, 
	       :postst => EXPR_BEG)
    #assert_tok(" /a/", :tREGEXP, "/a/", :prest => EXPR_ARG)
    #assert_tok("/a/", "/", "/", :prest => EXPR_ARG, :postst => EXPR_BEG)
    #assert_tok(" / ", "/", "/", :prest => EXPR_ARG, :postst => EXPR_BEG)
  end

  def test_38_left_brace
    assert_tok("{",     :tLBRACE, "{", 
	       :prest => EXPR_STATES - [EXPR_END, EXPR_ARG],
	       :postst => EXPR_BEG)
    assert_tok("{",     "{", "{", 
	       :prest => [EXPR_END, EXPR_ARG], :postst => EXPR_BEG)
  end

  def test_39_right_brace
    assert_tok("}",  "}",  "}", :prest => EXPR_STATES, :postst => EXPR_END)
  end

  def test_40_left_bracket
    assert_tok("[", "[", "[", :prest => [EXPR_FNAME, EXPR_DOT]) 
    assert_tok("[", :tLBRACK, "[", :prest => [EXPR_BEG, EXPR_MID],  :postst => EXPR_BEG) 
    assert_tok("[", "[", "[", 
	       :prest => EXPR_STATES - FNAME_AND_DOT - BEG_AND_MID-[EXPR_ARG],
	       :postst => EXPR_BEG) 
		assert_tok(" [", :tLBRACK, "[", :state_mode => EXPR_ARG, :state_pos => 1) 
  end

  def test_41_right_bracket
    assert_tok("]",  "]",  "]", :prest => EXPR_STATES, :postst => EXPR_END)
  end

  def test_42_aref
    assert_tok("[]", :tAREF, "[]", :prest => EXPR_FNAME, :postst => EXPR_FNAME)
    assert_tok("[]",  :tAREF, "[]", :prest => EXPR_DOT, :postst => EXPR_DOT)
    assert_tok("[]=", :tLBRACK, "[", :prest => EXPR_BEG)
  end

  def test_43_aset
    assert_tok("[]=",   :tASET, "[]=", :prest => EXPR_FNAME, :postst => EXPR_FNAME)
    assert_tok("[]=",   :tASET, "[]=", :prest => EXPR_DOT, :postst => EXPR_DOT)
    assert_tok("[]=", :tLBRACK, "[", :prest => EXPR_BEG)
  end

  def test_44_left_parenthesis # RACK
    assert_tok("(", :tLPAREN, "(", :prest => [EXPR_BEG, EXPR_MID], 
	       :postst => EXPR_BEG)
    assert_tok("(", "(", "(", :prest => EXPR_STATES - [EXPR_BEG, EXPR_MID],
	       :postst => EXPR_BEG)
  end
	
  def test_45_right_parenthesis # RACK
    assert_tok(")",  ")",  ")", :prest => EXPR_STATES, :postst => EXPR_END)
  end
	
  def test_46_string_single_quote
    assert_tok("'a'", :tSTRING, "a", :postst => EXPR_END)
    assert_tok("'a '", :tSTRING, "a ", :postst => EXPR_END)

    # Temporarily only raw strings
    assert_tok("'a\\''", :tSTRING, "a\\'", :postst => EXPR_END)
    #assert_tok("'a\\''", :tSTRING, "a'", :postst => EXPR_END)

    assert_tok("'a backslash \\'\\\\\\''", :tSTRING, 
	       "a backslash \\'\\\\\\\'",  :postst => EXPR_END)
    #assert_tok("'a backslash \\'\\\\\\''", :tSTRING, "a backslash '\'", 
    #           :postst => EXPR_END)

    assert_tok("'", "'", "'", :prest => [EXPR_FNAME, EXPR_DOT], 
	       :postst => EXPR_END)
  end

  def test_47_string_double_quote
    assert_tok('"a"', :tDSTRING, "a", :postst => EXPR_END)
    assert_tok('"a\t"', :tDSTRING, "a\\t", :postst => EXPR_END)
    assert_tok('"a\""', :tDSTRING, "a\\\"", :postst => EXPR_END)
    assert_tok(' "\111"', :tDSTRING, "\\111", :postst => EXPR_END)

    # Temporary handling of interleaved expressions:
    assert_tok('"a #{nil}"', :tDSTRING, 'a #{nil}', :postst => EXPR_END)
  end

  def test_48_string_here_document
    assert_tok("<<EOS\na\nEOS", :tDSTRING, "a\n", 
    	       :prest => EXPR_STATES - [EXPR_END, EXPR_CLASS, EXPR_ARG],
	       :postst => EXPR_END)
    assert_tok(" <<EOS\na\nEOS", :tDSTRING, "a\n", 
	       :prest => EXPR_ARG, :postst => EXPR_END)

    #indent neutral here document <<-token
    assert_tok("<<-EOS\na\n  EOS", :tDSTRING, "a\n", :postst => EXPR_END)

    #single quoted here document <<'token'
    assert_tok("<<'EOS'\na\nEOS", :tSTRING, "a\n", :postst => EXPR_END)

    #single quoted here document not replacing #{...}
    assert_tok("<<'EOS'\n\#{a}\nEOS", :tSTRING, "\#{a}\n", :postst => EXPR_END)

    # check that it leaves the newline chars
    assert_tok(" <<EOS\na\nEOS\n", :tDSTRING, "a\n", 
	       :prest => EXPR_ARG, :postst => EXPR_END, :postpos => 12)
  end

  def test_49_string_percent_q
    assert_tok("%q{a}", :tSTRING, "a", :postst => EXPR_END)
    assert_tok("%q<a >", :tSTRING, "a ", :postst => EXPR_END)
    assert_tok("%q[a']", :tSTRING, "a'", :postst => EXPR_END)
    assert_tok("%q(a\')", :tSTRING, "a'", :postst => EXPR_END)
    assert_tok("%q/a\\/", :tSTRING, "a\\", :postst => EXPR_END)
    assert_tok("%q{embedded {brackets}}", :tSTRING, "embedded {brackets}", 
	       :postst => EXPR_END)
  end

  def test_50_string_percent_Q
    assert_tok('%Q{a}', :tDSTRING, "a", :postst => EXPR_END)
    assert_tok('%Q<a\t>', :tDSTRING, "a\\t", :postst => EXPR_END)
    assert_tok('%Q[a\]', :tDSTRING, "a\\", :postst => EXPR_END)
    assert_tok('%Q( \111)', :tDSTRING, " \\111", :postst => EXPR_END)
    assert_tok('%Q| \111|', :tDSTRING, " \\111", :postst => EXPR_END)

    # Temporary handling of interleaved expressions:
    assert_tok('%Q{a #{nil}}', :tDSTRING, 'a #{nil}', :postst => EXPR_END)
  end

  def test_51_string_percent_delimited
    assert_tok('%<a>', :tDSTRING, "a", :postst => EXPR_END)
    assert_tok('%(a\t)', :tDSTRING, "a\\t", :postst => EXPR_END)
    assert_tok('%[a\"]', :tDSTRING, "a\\\"", :postst => EXPR_END)
    assert_tok('%{\111}', :tDSTRING, "\\111", :postst => EXPR_END)
    #assert_tok('%{ \111}', :tDSTRING, "\\111", :postst => EXPR_END)

    # Temporary handling of interleaved expressions:    
    assert_tok('%/a #{nil}/', :tDSTRING, 'a #{nil}', :postst => EXPR_END)
  end
	
  def test_52_regex_slash
    assert_tok("/a/", :tREGEXP, "a", :postst => EXPR_END)
    assert_tok("/=/", :tREGEXP, "=", :prest => [EXPR_BEG, EXPR_MID])
    assert_tok("/\Ahello/", :tREGEXP, "\Ahello", :postst => EXPR_END)
  end

  def test_53_regex_percent_r
    assert_tok("%r{a}", :tREGEXP, "a", :postst => EXPR_END)
    assert_tok("%r<Ahello>", :tREGEXP, "Ahello", :postst => EXPR_END)
  end

  #---------------------- Simple Statement Tests ---------------------------
	
  def test_simple_statement_1
    assert_tok_list( "22 + 22", 
		     [ [:tINTEGER, 22], ["+", "+"], [:tINTEGER, 22] ] )
  end
	
  def test_simple_statement_2
    assert_tok_list( "x=23", 
		     [ [:tIDENTIFIER, "x"], ["=","="], [:tINTEGER, 23] ] )
  end

  def test_simple_statement_3
    assert_tok_list( "class A\ndef m(a)\na + @val\nend\nend", 
		     [ [:kCLASS, "class"], [:tCONSTANT, "A"], ["\n", "\n"],
		       [:kDEF, "def"], [:tIDENTIFIER, "m"], ["(", "("],
		       [:tIDENTIFIER, "a"], [")", ")"], ["\n", "\n"],
		       [:tIDENTIFIER, "a"], ["+", "+"], [:tIVAR, "val"],
		       ["\n", "\n"], [:kEND, "end"],
		       ["\n", "\n"], [:kEND, "end"] ] )
  end

  #---------------------- Complex Statement Tests --------------------------

  def test_complex_statement_1
		string = <<-EOS
      def method1
		x=32
      end
    EOS
		assert_tok_list string,
      [ 
	#RF added newline tokens although unsure if they're correct?
	[:kDEF, "def"], [:tIDENTIFIER, "method1"], ["\n", "\n"],
	[:tIDENTIFIER, "x"], ["=","="], [:tINTEGER, 32], ["\n", "\n"],
	[:kEND, "end"], 
	["\n", "\n"] 
      ]
  end
  #-------------------- Assert Helper Functions -------------------------
	
  def assert_tok_list(string, tok_list)
    t = Lexer.new string
    t.state = EXPR_BEG
    tok_list.each do |token|
      parsed_token = t.next_token
      token.length.times do |i|
	if Token === parsed_token
	  if i == 0
	    assert_equals(token[i], parsed_token.token_type)
	  else
	    assert_equals(token[i], parsed_token[i-1])
	  end
	else
	  assert_equals(token[i], parsed_token[i])
	end
      end
    end
    assert_equals(false, t.next_token.token_type) #there should be no more tokens
  end
  
  def assert_position(position, expected = nil)
    assert_kind_of(Integer, position)
    assert_equals(expected, position) if expected
  end
  
  def assert_tok(string, tokenType, lexeme, restArgs = {})
    prests = restArgs[:prest]
    if not prests.kind_of?(Array)
      prests = [prests ? prests : EXPR_BEG]
    end
    prests.each do |pre_state|
      if restArgs[:state_mode] #handle moded state checks
	st = ModedState.new string
	st.mode = restArgs[:state_mode]
	st.position = restArgs[:state_pos] if restArgs[:state_pos]
	t = Lexer.new st 
      else # handle normal lexing operations
	t = Lexer.new string
	yield t if block_given?
	t.state = pre_state
      end
      token = t.next_token
      # Temporarily we allow both RaccToken and Token... (while converting
      # to token).
      unless RaccToken === token || Token === token
	assert(false, "Token is not RaccToken or Token")
      end
      assert_equals(tokenType, token.token_type)
      assert_equals(lexeme, token.lexeme)
      assert_position(token.position, restArgs[:pos])
      assert_position(t.position, restArgs[:postpos]) #if restArgs[:postpos]
      assert_equals(restArgs[:postst], t.state) if restArgs[:postst]
      assert_equals(restArgs[:space_seen], t.space_seen) if restArgs[:space_seen]
      assert_equals(restArgs[:info], token.info) if restArgs[:info]
      assert_equals(restArgs[:cond_stack], token.cond_stack) if restArgs[:cond_stack] # RACK - needed for parenthesis
      assert_equals(restArgs[:cond_nest], token.cond_nest) if restArgs[:cond_nest] # RACK (see above)
    end
  end
end
