# Unit tests for Ruby parser in rulator/parser
#
# Copyright (c) 2001, Robert Feldt, feldt@ce.chalmers.se
#
# $Id: utest_parser.rb,v 1.18 2001/12/17 00:30:01 feldt Exp $
#
require 'rubyunit'
require 'rulator/parser'
include Ruby
require 'rockit/term/match'

class RubyParserTest < TestCase	
  def setup
    @p = Parser.new
  end

  def test_01_initialize
    assert_kind_of(Parser, @p)
  end

  def test_02_alias_fitem_fitem
    assert_term(Body[Alias[SymbolLit["n"], SymbolLit["o"]]], 
		@p.parse("alias :n :o"))
  end

  def test_03_alias_gvar_gvar
    md = assert_term(Body[Alias[GlobalVar[:x], GlobalVar[:x]]], 
		     ast = @p.parse("alias $m $extend"))
    assert_equals("m",      md[1])
    assert_equals("extend", md[2])
  end

  def test_04_alias_gvar_backref
    md = assert_term(Body[Alias[GlobalVar[:x], BackRef[:x]]], 
		     @p.parse("alias $m $&"))
    assert_equals("m", md[1])
    assert_equals("&", md[2])
  end

  def test_05_plus
    assert_term(Body[Plus[IntegerLit[1], IntegerLit[9]]], 
		@p.parse("1   +9"))
  end

  def test_06_mult
    assert_term(Body[Mult[IntegerLit[2], IntegerLit[7]]], 
		@p.parse("2* 7"))
  end

  def test_07_div
    assert_term(Body[Div[IntegerLit[99], IntegerLit[3]]], 
		@p.parse("99/3"))
  end

  def test_08_operator_precedence
    # I'm not sure the Body created inside the parens should really be there.
    # Check it!
    assert_term(Body[Plus[Pow[IntegerLit[1], IntegerLit[2]], 
		          Mult[IntegerLit[2], 
		               Body[Minus[IntegerLit[3], 
			                  IntegerLit[2]]]]]],
		@p.parse("1**2 + 2 * (3 - 2)"))
  end

  def test_09_operator_associativity
    assert_term(Body[Minus[Minus[IntegerLit[17], IntegerLit[6]],
		           IntegerLit[2]]],
		@p.parse("17-6-2"))
  end

  def test_10_true_false_self_nil
    assert_term(Body[True[]],  @p.parse("true"))
    assert_term(Body[False[]], @p.parse("false"))
    assert_term(Body[Nil[]], @p.parse("nil"))
    assert_term(Body[Self[]], @p.parse("self"))
  end

  def test_11_symbols
    assert_term(Body[SymbolLit["simple"]], @p.parse(":simple"))
    assert_term(Body[SymbolLit["UPCASE"]], @p.parse(":UPCASE"))
    assert_term(Body[SymbolLit["+"]],      @p.parse(":+"))
    assert_term(Body[SymbolLit["[]"]],     @p.parse(":[]"))
    assert_term(Body[SymbolLit["<=>"]],     @p.parse(":<=>"))
  end

  def test_12_integer
    assert_term(Body[IntegerLit[10]],      @p.parse("10"))    
    assert_term(Body[IntegerLit[0x21]],    @p.parse("0x21"))    
    assert_term(Body[IntegerLit[0724]],    @p.parse("0724"))    
    assert_term(Body[IntegerLit[0b10_01]], @p.parse("0b10_01"))    
  end

  def test_13_float
    assert_term(Body[FloatLit[1.2]],       @p.parse("1.2"))
    assert_term(Body[FloatLit[.4_5e-3_5]], @p.parse(".4_5e-3_5"))
  end

  def test_14_array
    assert_term(Body[ArrayLit[[IntegerLit[1], IntegerLit[2]], nil]],
		@p.parse("[1,2]"))
  end

  def test_15_hash
    assert_term(Body[HashLit[Assoc[IntegerLit[1], True[]],
		             Assoc[False[],       FloatLit[0.4]]]],
		@p.parse("{1 => true, false => 0.4}"))
  end

  def test_16_variables
    assert_term(Body[GlobalVar["global"]], @p.parse("$global"))    
    assert_term(Body[InstanceVar["tage"]], @p.parse("@tage"))    
    assert_term(Body[ClassVar["mock"]],    @p.parse("@@mock"))    
  end

  def test_17_assignment
    assert_term(Body[Assign[Identifier["local"], IntegerLit[99_2]]],    
		@p.parse("local = 99_2"))    
  end

  def test_18_multiple_assignment
    assert_term(Body[Assign[Mlhs[Identifier["a"], Identifier["b"]], 
		            Mrhs[[FloatLit[.2], Self[]], nil]]],    
		@p.parse("a,b = .2,self"))
  end

  def test_19_if
    assert_term(Body[If[True[], 
		        Body[False[]],
		        If[Nil[],
		           Body[IntegerLit[4]],
		           Body[IntegerLit[5]]]]],
		@p.parse("if true then false elsif nil then 4 else 5 end"))
  end

  def test_20_if_mod
    assert_term(Body[If[False[], Assign[Identifier["a"], IntegerLit[0]]]],
		@p.parse("a = 0 if false"))
  end

  def test_21_unless
    assert_term(Body[Unless[True[], 
		            Body[False[]],
                            Body[IntegerLit[-1]]]],
		@p.parse("unless true then false else -1 end"))
  end

  def test_22_unless_mod
    assert_term(Body[Unless[False[], Assign[Identifier["a"], IntegerLit[0]]]],
		@p.parse("a = 0 unless false"))
  end

  def test_23_assign_to_attribute
    assert_term(Body[Assign[Access[Identifier["obj"], Identifier["state"]],
		            IntegerLit[10]]],
		@p.parse("obj.state = 10"))
  end

  def test_24_access_constant
    assert_term(Body[Access[Access[Constant["A"], Constant["B"]], 
		            Constant["C"]]],
		@p.parse("A::B::C"))
  end

  def test_25_method_call
    assert_term(Body[Call[Identifier["a"], Identifier["inspect"], nil]],
		@p.parse("a.inspect"))
  end

  def test_26_method_call_with_arg
    assert_term(Body[Call[Identifier["obj"], 
		          Identifier["plus"],
		          Args[[IntegerLit[1]]]]],
		@p.parse("obj.plus(1)"))
  end

  def test_27_method_call_with_args
    assert_term(Body[Call[Identifier["obj"], 
		          Identifier["plus"],
		          Args[[IntegerLit[1], True[]]]]],
		@p.parse("obj.plus(1, true)"))
  end

  def test_28_method_call_with_args_and_restarg
    assert_term(Body[Call[Identifier["o"], 
		          Identifier["b"],
		          Args[[Nil[]], nil, Identifier["a"]]]],
		@p.parse("o.b(nil, *a)"))
  end

  def test_29_method_call_with_args_assocs_and_restarg
    assert_term(Body[Call[Access[Constant["A"], Constant["B"]],
		          Identifier["c"],
		          Args[[Self[]], 
		               [Assoc[SymbolLit["opt1"], Constant["L"]]],
		               Identifier["a"]]]],
		@p.parse("A::B.c(self, :opt1 => L, *a)"))
  end

  def test_30_method_call_with_blockarg
    assert_term(Body[Call[nil, Identifier["m"],
		          Args[nil, nil, nil, Identifier["b"]]]],
		@p.parse("m(&b)"))
  end

  def atest_31_method_call_with_block
  end

  def test_32_empty_class
    assert_term(Body[DefClass[Constant["A"], nil, nil]],
		@p.parse("class A; end"))
  end

  def test_33_empty_subclass
    assert_term(Body[DefClass[Constant["A"], Constant["B"], nil]],
		@p.parse("class A < B; end"))
  end

  def test_34_simple_class
    ast = @p.parse <<-EOS
      class A < B
	def m(a)
	  a
	end
      end
    EOS
    assert_term(Body[DefClass[Constant["A"], Constant["B"], 
		              Body[DefMethod[Identifier["m"],
		                             Args[[Identifier["a"]]],
			                     Body[Identifier["a"]]]]]],
		ast)
  end

  def test_35_simple_class2
    s = <<-EOS
      class A < B
	def m(a)
	  a - @val
	end
      end
    EOS
    assert_term(Body[DefClass[Constant["A"], Constant["B"],
		              Body[DefMethod[Identifier["m"],
		                             Args[[Identifier["a"]]],
			                     Body[Minus[Identifier["a"],
			                                InstanceVar["val"]]]]]]], 
		@p.parse(s))
  end

  def test_36_while_modifier
    assert_term(Body[While[LessThan[Identifier["cnt"], IntegerLit[19]],
		           OpAssign[Identifier["cnt"], IntegerLit[1, nil], "+"]]],
		@p.parse("cnt += 1 while cnt < 19"))
  end

  def test_37_while
    s = <<-EOS
      while cnt >= 200
	cnt -= 7
      end
    EOS
    assert_term(Body[While[GreaterThanEq[Identifier["cnt"], IntegerLit[200]],
		           Body[OpAssign[Identifier["cnt"], 
			                 IntegerLit[7], "-"]]]],
		@p.parse(s))
  end

  def test_38_heredoc_followed_by_statement
    assert_term(Body[Assign[Identifier["a"], DStringLit[""]],
		     Assign[Identifier["b"], IntegerLit[2]]],
		@p.parse("a = <<EOS\nEOS\nb = 2"))
  end

  def test_39_module
    s = <<-EOS
      module RUNIT 
	TEST_CLASSES = {}

	def inherited(test_class)
	  TEST_CLASSES << test_class 
	end  
      end
    EOS
    assert_term(Body[
		  DefModule[
		    Constant["RUNIT"],
		    Body[Assign[Constant["TEST_CLASSES"], HashLit[]],
		      DefMethod[
			Identifier["inherited"],
			Args[[Identifier["test_class"]]],
			Body[
			  LeftShift[
			    Constant["TEST_CLASSES"], 
			    Identifier["test_class"]]]]]]],
		@p.parse(s))
  end

  def test_40_for
    s = <<-EOS
      sum = 0
      for i in (1..10).to_a
	sum += i
      end
    EOS
    assert_term(Body[Assign[Identifier["sum"], IntegerLit[0]],
		  For[Identifier["i"],
		    Call[Body[ARange[IntegerLit[1], IntegerLit[10], true]],
		      Identifier["to_a"], nil],
		    Body[
		      OpAssign[Identifier["sum"], Identifier["i"], "+"]]]],
		@p.parse(s))
  end

  def test_41_case
    s = <<-EOS
      case v
      when 1,3
	-20
      when A
	true
      else
	nil
      end
    EOS
  end

  def atest_42_aref
  end

  def atest_43_aset
    @p.parse("ary[0,2] = [3,5]")
  end

  def test_44_return
    s = <<-EOS
      def m
	return 1
      end
    EOS
    assert_term(Body[DefMethod[Identifier["m"], nil, 
		               Body[Return[Args[[IntegerLit[1]]]]]]],
		@p.parse(s))
  end

  def assert_term(expected, actual)
    assert_kind_of(Ruby::Term, actual)
    md = expected.match(actual)
    if md == nil
      matchdata = expected.match(actual, true)
      assert(false,
	     "Term mismatch: #{matchdata.mismatches.inspect}")
    end
    md
  end
end


