# Combinator library for building FormattedDocs - abstract descriptions of
# the layout of objects.
module Rockit
  module PrettyPrint
    # Abstract base class
    class FormattedDoc
      def to_formatted_doc
	self
      end

      def inspect_type
	self.type.inspect.split("::").last
      end

      alias inspect inspect_type
    end

    # A formatted string. Can be given names/types so that Writers can output
    # strings of different types in different ways. The convention is to use
    # symbols to denote types and subtypes.
    class Text < FormattedDoc
      attr_reader :string
      def initialize(aString, *types)
	super()
	@string, @types = aString, types
      end

      def inspect; @string.inspect; end
    end

    # Short hand constructor
    def text(aString, aName = nil); Text.new(aString, aName); end

    # Commonly used special cases when formatting source code
    def keyword(aString); Text.new(aString, :keyword); end
    def comment(aString); Text.new(aString, :comment); end

    # All short hand's call make_fdocs to convert non fdoc's to fdocs. Since
    # the to_format_doc call on String calls text you don't have to use text
    # if you don't need to give the string a type.
    def make_fdocs(docs)
      docs.map {|d| d.to_formatted_doc}
    end

    # Aggregated FDoc with at least one subdocument
    class CompoundFormattedDoc < FormattedDoc
      include Enumerable

      def initialize(*subDocs)
	unless subDocs.length > 0
	  raise ArgumentError, 
	    "Need at least one subdocument (#{self.type.inspect})" 
	end
	super()
	@sub_docs = subDocs
      end

      def each; @sub_docs.each {|d| yield(d)}; end

      def length; @sub_docs.length; end

      def inspect
	inspect_type + "(" +
	  @sub_docs.map {|d| d.inspect}.join(", ") +
	  ")"
	
      end
    end

    # Format the subdocs horizontally beside each other separated by a 
    # separator. The separator can be changed but defaults to "" ie. no
    # separation. If the subdocs does not fit on the line break and align 
    # the "spill" to have the same left margin as the first subdoc.
    #
    class Besides < CompoundFormattedDoc
      def initialize(*subDocs)
	@options = (Hash === subDocs.last) ? subDocs.pop : Hash.new
	super(*subDocs)
      end

      def separator()
	@options[:separator] || (@options[:separator] = "")
      end

      def separator=(newSeparator)
	@options[:separator] = newSeparator
      end
    end

    def besides(*subDocs); Besides.new(*make_fdocs(subDocs)); end

    # Format the subdocs horizontally with spaces between them. 
    def hor(*subDocs)
      Besides.new(*(make_fdocs(subDocs) + [{:separator => " "}]))
    end

    # Format the subdocs horizontally as elements of a list separated by
    # ", ". Note that we do *NOT* use Besides directly here since we want
    # elements to bind tightly to the separator coming after it. This
    # ensures that the split is between elements rather than between an 
    # element and its separator.
    def list(*subDocs)
      subDocs = subDocs[0..-2].map {|d| besides(d, ",")} << subDocs.last
      hor(*subDocs)
    end

    # Format subdocs vertically each one on top of the following. Left align
    # them.
    class Vertical < CompoundFormattedDoc; end

    def vert(*subDocs); Vertical.new(*make_fdocs(subDocs)); end

    # Indent a formatted doc. This only marks a FDoc as indented; its up to
    # the Writer dumping the FDoc to implement the indentation. For example,
    # PlainTextWriter by default uses an indent of "  ", ie. two spaces but
    # this can easily be changed to something else like four spaces or the 
    # tab.
    class Indent < FormattedDoc; end

    def indent(subDoc); Indent.new(subDoc.to_formatted_doc); end
  end
end

# Add a to_format_doc to all objects so that all objects can be formatted.
# It defaults to text(inspect) so this only works for simple objects. You'll
# have to override to give more approriate behavior in subclasses.
class Object
  def to_formatted_doc; text(inspect); end
end

class String
  def to_formatted_doc; text(self); end
end

class Array
  def to_formatted_doc
    besides("[", list(*self.map {|e| e.to_formatted_doc}), "]")
  end
end

class Hash
  def to_formatted_doc
    besides("{", list(*self.sort.map {|k,v| hor(k, "=>", v)}), "}")
  end
end
