# Basic class for Term's. Term's are used to represent sentences in a
# language. They represent nodes in abstract syntax trees.
#
# Copyright (c) 2001, Robert Feldt, feldt@ce.chalmers.se
#
# $Id: term.rb,v 1.9 2001/12/12 12:00:19 feldt Exp $
#
require 'graph/graphviz_dot'

module Rockit
  module ArefCreate
    def [](*args)
      self.new(*args)
    end
  end

  # Module with methods for creating sub-classes of Terms that have 
  # partially specified parameters to the super class's init method.
  # Classes wishing to add this feature should extend with this module.
  module PartialEvalSubClassFactory
    def new_sub_class(name,*childrenNames)
      cl = Class.new(self)
      cl.extend ArefCreate
      params = "@@params" + rand(1e10).inspect # Make it hard to guess
      cl.class_eval %{
       #{params} = []
       def self.inspect; #{params}[0]; end
       def self.children_names; #{params}[1]; end
      }
      childrenNames.map! {|n| to_string_name(n)}
      (cl.class_eval(params) << 
         to_string_name(name) << 
         (self.children_names + childrenNames)).freeze
      add_accessors(cl, childrenNames, self.children_names.length)
      cl
     end

    def add_accessors(klass, accessors, startIndex = 0)
      accessors.each_with_index do |name, index|
	check_name_validity(name)
        add_accessor(klass, name, startIndex + index)
      end
    end

    def add_accessor(klass, name, index)
      klass.class_eval %{
        def #{name}; self[#{index}]; end
        def #{name}=(newValue); self[#{index}] = newValue; end
      }      
    end

    def check_name_validity(name)
      unless name =~ /^[a-zA-Z][a-zA-Z0-9_]*$/
        raise ArgumentError, "No special chars (#{name}) in children names"
      end
      if Object.new.methods.include?(name)
        raise ArgumentError, "'#{name}' clashes with method in Object class"
      end
      if self.new.methods.include?(name)
        raise ArgumentError, "Name '#{name}' clashes with method in #{self.inspect} class"
      end    
    end

    def to_string_name(name)
      name = Symbol === name ? name.id2name : name
      raise unless String === name
      name
    end    
  end

  class Term
    extend PartialEvalSubClassFactory
    
    attr_reader :children

    def initialize(*values)
      @children = values
    end
  
    def ==(other)
      return false unless self.type == other.type
      @children.each_with_index {|c,i| return nil unless c == other[i]}
      true
    end

    def [](index)
      @children[index]
    end

    def []=(index, newValue)
      @children[index] = newValue
    end

    def inspect
      self.type.inspect + 
	"[" + @children.map {|v| v.inspect}.join(', ') + "]"
    end

    @@node_labeler = proc do |n|
      if Term === n
	n.type.inspect
      else
	s = n.inspect
	s = s[0...15] + "..." if Array == n.type && s.length > 15
	s
      end
    end

    @@node_shaper = proc do |n|
      Term === n ? "ellipse" : "box"
    end

    def to_graphics(filename, type = "png")
      dgp = add_to_dot
      dgp.node_labeler = @@node_labeler
      dgp.node_shaper = @@node_shaper
      puts dgp.to_dot_specification
      dgp.write_to_file(filename, type)
    end

    def Term.children_names
      []
    end

    def add_to_dot(dotGraphPrinter = DotGraphPrinter.new)
      @children.each_with_index do |value, index|
	if value
	  dotGraphPrinter.add_edge(self, value, 
				   self.type.children_names[index])
	end
	if Term === value
	  value.add_to_dot(dotGraphPrinter)
	elsif Array == value.type # Don't use === here since RaccToken is Array
	  value.each_with_index do |e, i| 
	    dotGraphPrinter.add_edge(value, value[i], i.to_s)
	    e.add_to_dot(dotGraphPrinter) if Term === e
	  end
	end
      end
      dotGraphPrinter
    end
  end

  class ArrayTerm < Term
    extend PartialEvalSubClassFactory
    def push(newValue)
      @children.push newValue
    end
  end
end
