require 'rubyunit'

require 'rockit/parse/std_library'
include Rockit::Parse

module TestGrammar
  extend Grammar

  language "TestLang"
  version  "1.0"

  Str = Letter

  start Str
end

class ParserLibraryTest < TestCase
  def assert_parse(string, parser)
    begin
      assert_equals(string, parser.parse(string))
    rescue Exception => e
      assert(false, "Does no match '#{string}'")
    end
  end

  @@valid_digits = ("0".."9").to_a 

  def test_01_digit
    @@valid_digits.each {|d| assert_parse(d, Digit)}
    assert_exception(NoMatchError) {Digit.parse "a"}
  end

  @@valid_hexdigits = ("0".."9").to_a + ("a".."f").to_a + ("A".."F").to_a

  def test_02_hexdigit
    @@valid_hexdigits.each {|d| assert_parse(d, HexDigit)}
    assert_exception(NoMatchError) {HexDigit.parse "g"}
  end

  @@valid_letters = ("a".."z").to_a + ("A".."Z").to_a

  def test_03_letter
    @@valid_letters.each {|l| assert_parse(l, Letter)}
    assert_exception(NoMatchError) {Letter.parse "1"}
  end

  def test_04_alphanum
    (@@valid_letters + @@valid_digits).each {|l| assert_parse(l, AlphaNum)}

    # This is ParseError and not NoMatchError since they have been
    # merged. FIX!
    #assert_exception(ParseError) {AlphaNum.parse "*"}
  end

  def test_05_newline_unix
    assert_parse("\n", NewlineUnix)
  end

  def test_06_newline_win
    assert_parse("\r\n", NewlineWin)
  end

  def test_07_restofline
    assert_parse("kjdshfdshfjhdskhfkdsf\r\n", RestOfLine)
    assert_parse("kjdshfdshfjhdskhfkdsf\n", RestOfLine)
  end

  def test_08_lines_until
    p = lines_until(skip(seq("EOS", Newline)))
    assert_equals("1\n2\n", p.parse("1\n2\nEOS\n"))
    assert_equals("1\r\n2\r\n", p.parse("1\r\n2\r\nEOS\r\n"))
  end

  def test_09_lines_until_with_partial_last_line
    p = lines_until(skip(seq("EOS", look_ahead(Newline))))
    s = OptimizedParseState.new "1\n2\nEOS\n"
    assert_equals("1\n2\n", p.parse(s))
    assert_equals("\n", s.current.chr)
    assert_equals(7, s.position)
  end

  def test_10_oring_strings
    p = "a" | "b"
    assert_kind_of(ChoiceParser, p)
    assert_equals("a", p.parse("a"))
    assert_equals("b", p.parse("b"))
  end

  def test_11_term
    t = term(:Int, :lexeme)
    assert(t.ancestors.include?(Rockit::Term))
    assert_equals(["lexeme"], t.children_names)

    t = term()
    assert_equals(Term, t)

    t = term(:Int, :_, :lexeme, :_)
    assert(t.ancestors.include?(Rockit::TermWithSkips))
    assert_equals(["lexeme"], t.children_names)
    assert_equals([0,2], t.___skip_indices___)
  end

  def test_12_into
    p = seq("a", "b") ^ (b = term(:B, :_, :b))
    assert_kind_of(Parser, p)
    ast = p.parse("ab")
    assert_kind_of(b, ast)
    assert_equals("b", ast.b)

    p = "a" & "b" ^ (b = term(:B, :_, :b)) \
      > "a" & "c" ^ (c = term(:C, :_, :c))
    assert_kind_of(Parser, p)
    ast = p.parse("ac")
    assert_kind_of(c, ast)
    assert_equals("c", ast.c)

    num = plus(Digit) ^ (number = term(:Number, :lexeme))
    num2 = plus(Digit).build(number)
    10.times do
      str = Array.new(1 + rand(10)).map {(?0 + rand(10)).chr}.join
      assert_equals(num2.parse(str), num.parse(str))
    end
  end

  def test_13_separator
    p1 = "a" & "b"
    p = p1 & "c" ^ term(:P, :a, :b)
    ws = mult(" " | "\t" | "\f" | "\v")
    p.set_separator( skip(maybe(ws)), [p1] )

    res = p.parse "ab c"
    assert_equals("ab", res.a)
    assert_equals("c", res.b)

    res = p.parse "ab\t\t c"
    assert_equals("ab", res.a)
    assert_equals("c", res.b)

    assert_exception(NoMatchError) { p.parse "a bc" }
  end

  def atest_14_left_recursive
    num = plus(Digit) ^ (number = term(:Number, :lexeme))
    ident = Upper & mult(Upper | Digit) ^ (identifier = term(:Identifier, :lexeme))
    operator = '+' | '-' | '*' | '/' | 'MOD'
    expr = recursive(
             ident > 
             skip("(") & recurse & skip(")") ^ term() > 
             recurse & operator & recurse ^
               (binExpr = term(:BinExpr, :left, :op, :right)) >
             num)

    assert_equals("123", expr.parse("123").lexeme)
    assert_equals("I", expr.parse("I").lexeme)
    ast = expr.parse("3+4")
    assert_kind_of(binExpr, ast)
    assert_equals("3", ast.left.lexeme)
    assert_equals("+", ast.op)
    assert_equals("4", ast.right.lexeme)
    #assert_equals("98", expr.parse("(98)").lexeme)
  end

  def test_15_grammar
    g = TestGrammar
    assert_equals("TestLang", g.language)
    assert_equals("a", g.parse("a"))
  end

  def test_16_nest
    p = nest( "/*", "*/" )
    assert_equals("/* A comment */", 
		  p.parse("/* A comment */"))
    assert_equals("/* A comment /* inner */ */", 
		  p.parse("/* A comment /* inner */ */"))
    assert_equals("/* A comment /* inner */ after */", 
		  p.parse("/* A comment /* inner */ after */"))
    assert_exception(NoMatchError) {p.parse "a"}
    assert_exception(UnexpectedEosError) {p.parse "/* a"}
  end

=begin
  # Test that repeated sequencing with operators does not
  # create redundant sequence parsers.
  # We should not enforce these tests since the req they embody does not
  # work when users hold onto intermediate parsers in order to mark them
  # as separated.
  def test_15_no_redundant_seq_parsers
    p = HexDigit & Digit & "c"
    assert_equals(3, p.sub_parsers.length)
  end

  def test_16_no_redundant_choice_parsers
    p = HexDigit | Digit | "c"
    assert_equals(3, p.sub_parsers.length)
  end
=end
end
