# Unit tests for rockit/term/term
#
# Copyright (c) 2001, Robert Feldt, feldt@ce.chalmers.se
#
# $Id: utest_term.rb,v 1.11 2001/12/12 17:48:30 feldt Exp $
#
require 'rubyunit'

require 'rockit/term/term'
require 'rockit/term/match'
include Rockit

module Ruby
  class Term < Rockit::Term
    extend PartialEvalSubClassFactory
    attr_accessor :pre_comments, :post_comments
  end
end

# Terms used in tests below
Body  = Term.new_sub_class(:Body,  :statements)
Alias = Term.new_sub_class(:Alias, :old, :new)
GlobalVar = Term.new_sub_class(:GlobalVar, :lexeme, :info)
BodyDirect = ArrayTerm.new_sub_class(:BodyDirect)

class TermTest < RUNIT::TestCase
  def test_01_new_term_class
    binaryExpression = Term.new_sub_class(:BinaryExpression, 
					  :left, "operator", "right")
    assert(binaryExpression.ancestors[1] == Term)
    assert_equals(["left", "operator", "right"], 
		  binaryExpression.children_names)
  end

  def test_02_erroneous_children_names
    # with special chars
    assert_exception(ArgumentError) {Term.new_sub_class(:T, :left?)} 
    assert_exception(ArgumentError) {Term.new_sub_class(:T, :op!)} 

    # same name as method in Object
    assert_exception(ArgumentError) {Term.new_sub_class(:T, :id)} 
    assert_exception(ArgumentError) {Term.new_sub_class(:T, :__send__)} 
  end

  def test_03_creation
    binaryExpression = Term.new_sub_class(:BinaryExpression, 
					  :left, "operator", "right")
    b1 = binaryExpression.new(1, :+, 2)
    assert_kind_of(binaryExpression, b1)
    assert_kind_of(Term, b1)
  end

  def test_04_accessor_methods
    binaryExpression = Term.new_sub_class(:BinaryExpression, 
					  :left, "operator", "right")
    b1 = binaryExpression.new(1, :+, 2)
    assert_equals(1,  b1.left)
    assert_equals(:+, b1.operator)
    assert_equals(2,  b1.right)
  end

  def test_05_aref
    binaryExpression = Term.new_sub_class(:BinaryExpression, 
					  :left, "operator", "right")
    b1 = binaryExpression.new(1, :+, 2)
    assert_equals(1,  b1[0])
    assert_equals(:+, b1[1])
    assert_equals(2,  b1[2])
  end

  def test_06_add_to_dot
    binaryExpression = Term.new_sub_class(:BinaryExpression, 
					  :left, "operator", "right")
    b1 = binaryExpression.new(1, :*, 2)
    b2 = binaryExpression.new(b1, :+, 4)
    dgp = b2.add_to_dot
    assert_kind_of(DotGraphPrinter, dgp)
  end

  def test_07_real_subclass_with_attributes
    al = Ruby::Term.new_sub_class(:Alias, :old, :new)
    assert(al.ancestors.include?(Rockit::Term))
    assert(al.ancestors.include?(Ruby::Term))

    a = al["a", "b"]
    assert_kind_of(Ruby::Term, a)
    assert(a.respond_to?(:pre_comments))
    assert(a.respond_to?(:post_comments))
    assert(a.respond_to?(:pre_comments=))
    assert(a.respond_to?(:post_comments=))
  end

  def test_08_subclassing_a_sublcass
    loop       = Ruby::Term.new_sub_class(:Loop, :body)
    while_loop = loop.new_sub_class(:While, :condition)

    assert_equals(["body", "condition"], while_loop.children_names)

    wl = while_loop[body = [], condition = true] # Dummy values
    assert_kind_of(while_loop,   wl)
    assert_kind_of(loop,         wl)
    assert_kind_of(Ruby::Term,   wl)
    assert_kind_of(Rockit::Term, wl)

    assert_equals(body,      wl.body)
    assert_equals(condition, wl.condition)

    wl.pre_comments = a = ["adada", "muppel"]
    assert_equals(a, wl.pre_comments)
    wl.post_comments = a = [1, 2]
    assert_equals(a, wl.post_comments)
  end

  def test_09_match
    t       = Body[Alias[1,2]]
    pattern = Body[Alias[:x, :x]]
    md = pattern.match(t)
    assert_kind_of(Term::MatchData, md)
    assert_equals(t, md[0])
    assert_equals(1, md[1])
    assert_equals(2, md[2])

    md2 = Body[:x].match(Alias[1,2])
    assert_equals(nil, md2)

    md3 = Body[Alias[1,:x]].match(Body[1])
    assert_equals(nil, md3)

    md4 = Body[Alias[1, :x]].match(Body[Alias[1, 2]])
    assert_equals(2, md4[1])
    
    md5 = Body[Alias["tage", "brage"]].match(Body[Alias["@tage", "@brage"]])
    assert_equals(nil, md5)
  end

  def test_11_aset
    binaryExpression = Term.new_sub_class(:BinaryExpression, 
					  :left, "operator", "right")
    b1 = binaryExpression.new(1, :+, 2)
    b1.left = 3
    assert_equals(3,  b1[0])
    assert_equals(:+, b1[1])
    assert_equals(2,  b1[2])
  end

  def test_12_add_to_dot_when_arry
    # A problem is that when Arrays are used to collect Terms into one
    # children of a parent Term the terms in the Array are not dumped to dot.
    # Best would be to use some slot scheme with one slot for each element.

    body = Term.new_sub_class(:Body, :statements)
    binaryExpression = Term.new_sub_class(:BinaryExpression, 
					  :left, "operator", "right")
    be = binaryExpression.new(1, :*, 2)
    b1 = body.new([be])
    dgp = b1.add_to_dot
    assert_kind_of(DotGraphPrinter, dgp)
    # On node for "Body", one for its array child, one for the BinExpr and
    # three for its children => 6 nodes. We cheat since there is no
    # way to access the nodes in the DotGraphPrinter. Change it!
    assert_equals(6, dgp.instance_eval("@nodes").length)
  end

  def test_13_push_children
    b = BodyDirect[1]
    b.push 2
    assert_equals(1, b[0])
  end

  def test_14_term_equslity
    assert_equals(Body[1], Body[1])
  end

  def test_15_match_retaining_mismatches
    md = Body[:x].match(Alias[1,2], true)
    assert_equals(1, md.mismatches.length)

    md2 = Body[1].match(Body[3], true)
    assert_equals(1, md2.mismatches.length)
    assert_equals([1,3], md2.mismatches.first) 
  end

=begin
  def test_13_subclass_with_prespecified_childrens
    token       = Term.new_sub_class(:Token, :token_type, :lexeme, :info)
    globalVar   = token.new_sub_class(:GlobalVar, [:tGVAR])
    gv = globalVar["$a", nil]
    assert_kind_of(token, gv)
    assert_kind_of(globalVar, gv)
    assert_equals(:tGVAR, gv.token_type)
    assert_equals("$a", gv.lexeme)
    assert_equals(nil, gv.info)    
  end
=end


end
